import logging
import re
from pathlib import Path

import click

from youte.exceptions import InvalidFileName

logger = logging.getLogger(__name__)


def validate_file(file_name: str, suffix=None):
    path = Path(file_name)
    if suffix and path.suffix != suffix:
        raise InvalidFileName("File name must end in `%s`." % suffix)
    return path


def check_file_overwrite(file_path: Path) -> Path:
    if file_path.exists():
        flag = click.confirm(f"{file_path} already exists. Keep writing to it?")
        if flag:
            logger.info(f"Updating existing file {file_path.resolve()}...")
        else:
            file_name = input("Input new file name: ")
            file_path = Path(file_name)

    return file_path


def create_utc_datetime_string(string: str):
    return f"{string}T00:00:00Z"


def validate_date_string(string: str) -> bool:
    pattern = r"^([0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])$"
    if re.match(pattern, string):
        return True
    else:
        return False



import csv
import html
import json
import logging
import re
import sqlite3
from pathlib import Path
from typing import Iterable, Mapping, Sequence, Union

import click
from tqdm import tqdm

import youte.table_mappings as mappings

logger = logging.getLogger()


def connect_db(db_path: Union[str, Path]) -> sqlite3.Connection:
    """Helper function for opening DB connections with standard config set."""
    return sqlite3.connect(db_path, isolation_level=None)


def master_tidy(filepath: str, output: Union[str, Path]) -> None:
    with open(filepath, mode="r") as file:
        first_response = json.loads(file.readline())
        kind = first_response["kind"]

    if "video" in kind:
        logger.info(f"{kind}")
        tidy_video(filepath, output)
    elif "channel" in kind:
        tidy_channel(filepath, output)
    elif "comment" in kind:
        tidy_comments(filepath, output)
    elif "search" in kind:
        tidy_search(filepath, output)


def get_items(filepath: str) -> Iterable[dict]:
    with open(filepath, mode="r") as file:
        responses = (row.rstrip() for row in file.readlines())
        for response in responses:
            try:
                items = json.loads(response)["items"]
                for item in items:
                    yield item
            except KeyError:
                logger.warning("No items found in JSON response")


def _unnest_map(filepath: Mapping) -> Iterable:
    for key, value in filepath.items():
        if isinstance(value, Mapping):
            yield from _annotate_key(value, name=key)
        else:
            yield key, value


def _annotate_key(filepath, name):
    for key, value in filepath.items():
        if isinstance(value, Mapping):
            yield from _annotate_key(value, name=key)
        else:
            yield f"{name}_{key}", value


def _normalise_name(string):
    pattern = re.compile(r"([A-Z])")
    new_string = re.sub(pattern, r"_\1", string)
    return new_string.lower()


def tidy_search(filepath: str, output: Union[str, Path]) -> None:
    db = connect_db(output)

    db.execute("BEGIN")
    db.execute(
        """
        CREATE TABLE IF NOT EXISTS search_results(
             id primary key,
             kind,
             published_at,
             title,
             description,
             thumbnails,
             channel_title,
             channel_id,
             live_broadcast_content
             );
        """
    )
    items = get_items(filepath=filepath)
    total = 0

    for item in tqdm(items):
        kind = item["id"]["kind"]
        video_id = item["id"]["videoId"]
        snippet = item["snippet"]
        channel_id = snippet["channelId"]
        published_at = snippet["publishedAt"]
        title = html.unescape(snippet["title"])
        description = snippet["description"]
        thumbnail = json.dumps(snippet["thumbnails"])
        channel_title = snippet["channelTitle"]
        live_broadcast_content = snippet["liveBroadcastContent"]

        db.execute(
            """
            REPLACE INTO search_results
            VALUES (?,?,?,?,?,?,?,?,?)
            """,
            (
                video_id,
                kind,
                published_at,
                title,
                description,
                thumbnail,
                channel_title,
                channel_id,
                live_broadcast_content,
            ),
        )

        total += 1

    db.execute("commit")

    db.close()
    click.echo(f"{total} items processed.")
    click.echo(f"Data stored in {output} in `search_results` table.")


def _get_mapping(item: dict, resource_kind) -> dict:
    valid_kinds = ["search", "channels", "videos", "comments"]

    if resource_kind not in valid_kinds:
        raise ValueError(f"resource_kind must be one of {str(valid_kinds)}")

    mapping = dict()

    if resource_kind == "search":
        mapping["kind"] = item["id"]["kind"]
        for elem in item["id"]:
            if "id" in elem or "Id" in elem:
                mapping["id"] = item["id"][elem]
        snippet = item["snippet"]
        mapping["channel_id"] = snippet["channelId"]
        mapping["published_at"] = snippet["publishedAt"]
        mapping["title"] = html.unescape(snippet["title"])
        mapping["description"] = snippet["description"]
        mapping["thumbnail"] = json.dumps(snippet["thumbnails"])
        mapping["channel_title"] = snippet["channelTitle"]
        mapping["live_broadcast_content"] = snippet["liveBroadcastContent"]

    if resource_kind == "comments":
        if "topLevelComment" in item["snippet"]:
            snippet = item["snippet"]["topLevelComment"]["snippet"]
            mapping["reply_count"] = item["snippet"].get("totalReplyCount")
            mapping["can_reply"] = item["snippet"].get("canReply")
        else:
            mapping["reply_count"] = None
            mapping["can_reply"] = None
            snippet = item["snippet"]

        mapping["comment_id"] = item["id"]
        mapping["video_id"] = snippet.get("videoId")
        mapping["channel_id"] = snippet.get("channelId")
        mapping["parent_id"] = snippet.get("parentId")
        mapping["text_display"] = html.unescape(snippet["textDisplay"])
        mapping["text_original"] = snippet["textOriginal"]
        mapping["author_name"] = snippet["authorDisplayName"]
        mapping["author_channel_url"] = snippet["authorChannelUrl"]
        mapping["author_channel_id"] = snippet["authorChannelId"]["value"]
        mapping["can_rate"] = snippet["canRate"]
        mapping["viewer_rating"] = snippet["viewerRating"]
        mapping["like_count"] = snippet["likeCount"]
        mapping["published_at"] = snippet["publishedAt"]
        mapping["updated_at"] = snippet["updatedAt"]

    if resource_kind == "videos":
        snippet = item["snippet"]
        content_details = item["contentDetails"]
        status = item["status"]
        statistics = item["statistics"]
        topic_details = item.get("topicDetails")
        content_rating = content_details["contentRating"]
        recording = item["recordingDetails"]
        region_restriction = content_details.get("regionRestriction")

        mapping["video_id"] = item["id"]
        mapping["published_at"] = snippet["publishedAt"]
        mapping["channel_id"] = snippet["channelId"]
        mapping["title"] = snippet["title"]
        mapping["description"] = snippet["description"]
        mapping["thumbnails"] = json.dumps(snippet["thumbnails"])
        mapping["channel_title"] = snippet["channelTitle"]
        mapping["category_id"] = snippet["categoryId"]
        mapping["live_broadcast_content"] = snippet["liveBroadcastContent"]
        mapping["localized_title"] = snippet["localized"]["title"]
        mapping["localized_description"] = snippet["localized"]["description"]
        mapping["duration"] = content_details["duration"]
        mapping["dimension"] = content_details["dimension"]
        mapping["definition"] = content_details["definition"]
        mapping["caption"] = content_details["caption"]
        mapping["licensed_content"] = content_details["licensedContent"]
        mapping["upload_status"] = status["uploadStatus"]
        mapping["privacy_status"] = status["privacyStatus"]
        mapping["license"] = status["license"]
        mapping["public_stats_viewable"] = status["publicStatsViewable"]
        mapping["made_for_kids"] = status["madeForKids"]

        # optional properties
        mapping["default_language"] = snippet.get("defaultLanguage")
        mapping["default_audio_language"] = snippet.get("defaultAudioLanguage")
        mapping["rejection_reason"] = status.get("rejectionReason")
        mapping["region_allowed"] = (
            str(region_restriction.get("allowed")) if region_restriction else None
        )
        mapping["region_blocked"] = (
            str(region_restriction.get("blocked")) if region_restriction else None
        )
        mapping["yt_rating"] = content_rating.get("ytRating")
        mapping["tags"] = str(snippet.get("tags"))
        mapping["view_count"] = statistics.get("viewCount")
        mapping["like_count"] = statistics.get("likeCount")
        mapping["comment_count"] = statistics.get("commentCount")
        mapping["topic_categories"] = (
            str(topic_details["topicCategories"]) if topic_details else None
        )
        mapping["recording_location"] = str(recording.get("location"))

    if resource_kind == "channels":
        snippet = item["snippet"]
        content_details = item["contentDetails"]
        related_playlists = content_details.get("relatedPlaylists")
        status = item["status"]
        statistics = item["statistics"]
        topic_details = item.get("topicDetails")
        branding = item.get("brandingSettings")
        owner = item["contentOwnerDetails"]

        mapping["channel_id"] = item["id"]
        mapping["title"] = snippet["title"]
        mapping["description"] = snippet["description"]
        mapping["published_at"] = snippet["publishedAt"]
        mapping["custom_url"] = snippet.get("customUrl")
        mapping["thumbnails"] = json.dumps(snippet["thumbnails"])
        mapping["default_language"] = snippet.get("defaultLanguage")
        mapping["country"] = snippet.get("country")
        mapping["localized_title"] = snippet["localized"]["title"]
        mapping["localized_description"] = snippet["localized"]["description"]
        mapping["related_playlists_likes"] = (
            related_playlists.get("likes") if related_playlists else None
        )
        mapping["related_playlists_uploads"] = (
            related_playlists.get("uploads") if related_playlists else None
        )
        mapping["view_count"] = statistics.get("viewCount")
        mapping["subscriber_count"] = statistics.get("subscriberCount")
        mapping["video_count"] = statistics.get("videoCount")
        mapping["hidden_subscriber_count"] = statistics["hiddenSubscriberCount"]
        mapping["topic_ids"] = (
            str(topic_details.get("topicIds")) if topic_details else None
        )
        mapping["topic_categories"] = (
            str(topic_details.get("topicCategories")) if topic_details else None
        )
        mapping["privacy_status"] = status["privacyStatus"]
        mapping["is_linked"] = status.get("isLinked")
        mapping["made_for_kids"] = status.get("madeForKids")
        mapping["keywords"] = branding["channel"].get("keywords")
        mapping["moderate_comments"] = branding["channel"].get("moderateComments")
        mapping["unsubscribed_trailer"] = branding["channel"].get("unsubscribedTrailer")
        mapping["content_owner"] = owner.get("contentOwner")
        mapping["time_linked"] = owner.get("timeLinked")

    return mapping


def _write_csv(source: Sequence[Mapping], outfile: Union[str, Path]) -> None:
    fieldnames = source[0].keys()

    with open(outfile, "a", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=fieldnames)

        if f.tell() == 0:
            writer.writeheader()
        writer.writerows(source)


def tidy_to_csv(
    items: Iterable[Mapping], output: Union[str, Path], resource_kind: str
) -> None:
    to_write = []

    for item in tqdm(items):
        mapping = _get_mapping(item, resource_kind=resource_kind)
        to_write.append(mapping)

    _write_csv(source=to_write, outfile=output)


def tidy_video(filepath: str, output: Union[str, Path]) -> None:
    db = connect_db(output)

    db.execute("begin")
    db.execute(mappings.VIDEO_SQL_TABLE["create"])
    db.execute("commit")

    items = get_items(filepath=filepath)
    total = 0

    for item in tqdm(items):
        video_mapping = _get_mapping(item, "videos")

        db.execute("BEGIN")
        db.execute(mappings.VIDEO_SQL_TABLE["insert"], video_mapping)
        db.execute("COMMIT")

        total += 1

    db.close()
    click.echo(f"{total} items processed.")
    click.echo(f"Data stored in {output} in `videos` table.")


def tidy_channel(filepath: str, output: Union[str, Path]) -> None:
    db = connect_db(output)

    db.execute("begin")
    db.execute(mappings.CHANNEL_SQL_TABLE["create"])
    db.execute("commit")

    items = get_items(filepath=filepath)
    total = 0

    for item in tqdm(items):
        channel_mapping = _get_mapping(item, "channels")

        snippet = item["snippet"]
        content_details = item["contentDetails"]
        related_playlists = content_details.get("relatedPlaylists")
        status = item["status"]
        statistics = item["statistics"]
        topic_details = item.get("topicDetails")
        branding = item.get("brandingSettings")
        owner = item["contentOwnerDetails"]

        channel_mapping["channel_id"] = item["id"]
        channel_mapping["title"] = snippet["title"]
        channel_mapping["description"] = snippet["description"]
        channel_mapping["published_at"] = snippet["publishedAt"]
        channel_mapping["custom_url"] = snippet.get("customUrl")
        channel_mapping["thumbnails"] = json.dumps(snippet["thumbnails"])
        channel_mapping["default_language"] = snippet.get("defaultLanguage")
        channel_mapping["country"] = snippet.get("country")
        channel_mapping["localized_title"] = snippet["localized"]["title"]
        channel_mapping["localized_description"] = snippet["localized"]["description"]
        channel_mapping["related_playlists_likes"] = (
            related_playlists.get("likes") if related_playlists else None
        )
        channel_mapping["related_playlists_uploads"] = (
            related_playlists.get("uploads") if related_playlists else None
        )
        channel_mapping["view_count"] = statistics.get("viewCount")
        channel_mapping["subscriber_count"] = statistics.get("subscriberCount")
        channel_mapping["video_count"] = statistics.get("videoCount")
        channel_mapping["hidden_subscriber_count"] = statistics["hiddenSubscriberCount"]
        channel_mapping["topic_ids"] = (
            str(topic_details.get("topicIds")) if topic_details else None
        )
        channel_mapping["topic_categories"] = (
            str(topic_details.get("topicCategories")) if topic_details else None
        )
        channel_mapping["privacy_status"] = status["privacyStatus"]
        channel_mapping["is_linked"] = status.get("isLinked")
        channel_mapping["made_for_kids"] = status.get("madeForKids")
        channel_mapping["keywords"] = branding["channel"].get("keywords")
        channel_mapping["moderate_comments"] = branding["channel"].get(
            "moderateComments"
        )
        channel_mapping["unsubscribed_trailer"] = branding["channel"].get(
            "unsubscribedTrailer"
        )
        channel_mapping["content_owner"] = owner.get("contentOwner")
        channel_mapping["time_linked"] = owner.get("timeLinked")

        db.execute("begin")
        db.execute(mappings.CHANNEL_SQL_TABLE["insert"], channel_mapping)
        db.execute("commit")

        total += 1

    db.close()
    click.echo(f"{total} items processed.")
    click.echo(f"Data stored in {output} in `channels` table.")


def tidy_comments(filepath, output: Union[str, Path]) -> None:
    db = connect_db(output)

    db.execute("begin")
    db.execute(mappings.COMMENT_SQL_TABLE["create"])
    db.execute("commit")

    items = get_items(filepath=filepath)
    total = 0

    for item in tqdm(items):
        comment_mapping = _get_mapping(item, "comments")

        db.execute("BEGIN")
        db.execute(mappings.COMMENT_SQL_TABLE["insert"], comment_mapping)
        db.execute("COMMIT")

        total += 1

    db.close()
    click.echo(f"{total} items processed.")
    click.echo(f"Data stored in {output} in `comments` table.")


def get_id(items: Iterable[dict]) -> Iterable[str]:
    for item in items:
        id_ = item["id"]
        if isinstance(id_, dict):
            for elem in id_:
                if "id" in elem or "Id" in elem:
                    id_ = id_[elem]
        yield id_



VIDEO_SQL_TABLE = {
    "create": """
        CREATE TABLE IF NOT EXISTS videos (
            video_id PRIMARY KEY,
            published_at,
            channel_id,
            title,
            description,
            thumbnails,   -- dictionary of thumbnail versions
            channel_title,
            tags,
            category_id,
            live_broadcast_content,
            default_language,
            localized_title,
            localized_description,
            default_audio_language,
            duration,
            dimension,
            definition,  -- high or standard definition video
            caption,
            licensed_content,
            region_allowed,
            region_blocked,
            yt_rating,  -- age-restricted content
            upload_status,
            rejection_reason,
            privacy_status,
            license,
            public_stats_viewable,
            made_for_kids,
            view_count,
            like_count,
            comment_count,
            topic_categories,
            recording_location
        )
              """,
    "insert": """
        INSERT OR REPLACE INTO videos
        VALUES (
            :video_id,
            :published_at,
            :channel_id,
            :title,
            :description,
            :thumbnails,
            :channel_title,
            :tags,
            :category_id,
            :live_broadcast_content,
            :default_language,
            :localized_title,
            :localized_description,
            :default_audio_language,
            :duration,
            :dimension,
            :definition,
            :caption,
            :licensed_content,
            :region_allowed,
            :region_blocked,
            :yt_rating,
            :upload_status,
            :rejection_reason,
            :privacy_status,
            :license,
            :public_stats_viewable,
            :made_for_kids,
            :view_count,
            :like_count,
            :comment_count,
            :topic_categories,
            :recording_location
        )
              """,
}

CHANNEL_SQL_TABLE = {
    "create": """
        CREATE TABLE IF NOT EXISTS channels(
            channel_id PRIMARY KEY,
            title,
            description,
            custom_url,
            published_at,
            thumbnails, -- dictionary of thumbnail versions
            default_language,
            localized_title,
            localized_description,
            country,
            related_playlists_likes, -- ID of the playlist containing
                                     -- the channel's liked videos
            related_playlists_uploads, -- playlist of channel's uploaded videos
            view_count,
            subscriber_count,
            hidden_subscriber_count,
            video_count,
            topic_ids, -- list of topic ids
            topic_categories, -- list of topic categories
            privacy_status,
            is_linked, -- channel data identifies user already linked with YT
            made_for_kids,
            keywords,
            moderate_comments,
            unsubscribed_trailer,
            content_owner,
            time_linked
            )
              """,
    "insert": """
        INSERT OR REPLACE INTO channels
        VALUES (
            :channel_id,
            :title,
            :description,
            :custom_url,
            :published_at,
            :thumbnails,
            :default_language,
            :localized_title,
            :localized_description,
            :country,
            :related_playlists_likes,
            :related_playlists_uploads,
            :view_count,
            :subscriber_count,
            :hidden_subscriber_count,
            :video_count,
            :topic_ids,
            :topic_categories,
            :privacy_status,
            :is_linked,
            :made_for_kids,
            :keywords,
            :moderate_comments,
            :unsubscribed_trailer,
            :content_owner,
            :time_linked
            )
              """,
}

COMMENT_SQL_TABLE = {
    "create": """
        CREATE TABLE IF NOT EXISTS comments(
            comment_id PRIMARY KEY,
            video_id,
            channel_id,
            parent_id,
            can_reply,
            reply_count,
            text_display,
            text_original,
            author_name,
            author_channel_url,
            author_channel_id,
            can_rate,
            viewer_rating,
            like_count,
            published_at,
            updated_at
            )
             """,
    "insert": """
        INSERT OR REPLACE INTO comments
        VALUES (
            :comment_id,
            :video_id,
            :channel_id,
            :parent_id,
            :can_reply,
            :reply_count,
            :text_display,
            :text_original,
            :author_name,
            :author_channel_url,
            :author_channel_id,
            :can_rate,
            :viewer_rating,
            :like_count,
            :published_at,
            :updated_at
            )
            """,
}

SEARCH_SQL_TABLE = {
    "create": """
        CREATE TABLE IF NOT EXISTS search_results(
             id primary key,
             kind,
             published_at,
             title,
             description,
             thumbnails,
             channel_title,
             channel_id,
             live_broadcast_content
             );
        """,
    "insert": """
        INSERT OR REPLACE INTO search_results
        VALUES (
             :id,
             :kind,
             :published_at,
             :title,
             :description,
             :thumbnails,
             :channel_title,
             :channel_id,
             :live_broadcast_content
             );
        """,
}



import os
import pytest
from click.testing import CliRunner
from pathlib import Path

from youte.cli import youte

API_KEY = os.environ["STAGING_API_KEY"]
NAME = "tester"


@pytest.fixture()
def runner() -> CliRunner:
    runner = CliRunner()
    yield runner
    try:
        os.remove("youte.log")
    except FileNotFoundError:
        pass


def test_hydrate_no_input(runner):
    results = runner.invoke(youte, ["hydrate"])
    assert results.exception


@pytest.fixture()
def hydrate_args() -> list:
    return ["hydrate", "--key", API_KEY]


@pytest.mark.parametrize(
    "extra_args",
    [
        ["FZCWXlJWueM"],
        ["FZCWXlJWueM", "9nvdawlwATQ", "XH9-Dx_RwDw"],
        ["--file-path", Path("tests") / "video_ids.csv"],
    ],
    ids=["one_id", "multi_ids", "from_file"],
)
def test_hydrate_video(runner, hydrate_args, extra_args):
    results = runner.invoke(youte, hydrate_args + extra_args)
    assert results.exit_code == 0
    assert "youtube#videoListResponse" in results.output
    assert "title" in results.output


def test_hydrate_channel(runner, hydrate_args):
    results = runner.invoke(
        youte,
        hydrate_args + ["-f", Path("tests") / "channel_ids.csv", "--kind=channels"],
    )
    assert results.exit_code == 0
    assert "youtube#channelListResponse" in results.output
    assert "title" in results.output


@pytest.fixture()
def comment_args() -> list:
    return ["get-comments", "--key", API_KEY]


@pytest.mark.parametrize(
    "extra_args",
    [
        ["-v", "XmYnsO7iSI8"],
        ["-v", "XmYnsO7iSI8", "4oqjcKenCH8", "_b6NgY_pMdw"],
        ["-v", "--file-path", Path("tests") / "ids_to_get_comment.csv"],
    ],
    ids=["one_id", "multi_ids", "from_file"],
)
def test_get_comments_on_video(runner, comment_args, extra_args):
    results = runner.invoke(youte, comment_args + extra_args)
    assert results.exit_code == 0
    assert "youtube#commentThreadListResponse" in results.output
    assert "topLevelComment" in results.output


def test_dehydrate(runner):
    results = runner.invoke(youte, ["dehydrate", "tests/search_results.json"])
    assert results.exit_code == 0
    assert "AYdQvnWtEHU" in results.output


def test_dehydrate_fail(runner):
    results = runner.invoke(youte, ["dehydrate", "tests/channel_ids.csv"])
    assert results.exception



import os
import json
import pytest
from click.testing import CliRunner
from pathlib import Path

from youte.config import YouteConfig
from youte.collector import Youte
from youte.cli import youte


API_KEY = os.environ["STAGING_API_KEY"]
NAME = "tester"


@pytest.fixture()
def config_path(tmp_path):
    config_name = "test_config"
    return tmp_path / config_name


def test_config_empty(config_path):
    youte_conf = YouteConfig(filename=str(config_path))
    assert not config_path.exists()


def test_config_add(config_path):
    youte_conf = YouteConfig(filename=str(config_path))
    youte_conf.add_profile(NAME, API_KEY)

    assert config_path.exists()


@pytest.fixture()
def youte_obj(config_path) -> Youte:
    youte_obj = Youte(api_key=API_KEY)
    yield youte_obj
    if os.path.exists(".youte.history"):
        os.rmdir(".youte.history")


@pytest.fixture()
def runner() -> CliRunner:
    runner = CliRunner()
    yield runner
    try:
        os.remove("youte.log")
    except FileNotFoundError:
        pass


def test_empty_youte(youte_obj):
    assert youte_obj.api_key == API_KEY


@pytest.fixture()
def search_params() -> dict:
    params = {
        "standard-no-output": [
            "search",
            "harry potter",
            "--from",
            "2021-01-01",
            "--key",
            API_KEY,
            "--limit",
            "2"
        ],
        "standard-output": [
            "search",
            "harry potter",
            "--from",
            "2021-01-01",
            "--to",
            "2021-01-02",
            "--key",
            API_KEY,
            "--output",
            "output.json",
            "--limit",
            "2"
        ],
        "wrong-date-format": [
            "search",
            "harry potter",
            "--from",
            "10-01-2021",
            "--to",
            "2021-02-01",
            "--key",
            API_KEY,
            "--limit",
            "2"
        ],
        "ordered-by-relevance": [
            "search",
            "harry potter",
            "--order",
            "relevance",
            "--key",
            API_KEY,
            "--limit",
            "2"
        ],
        "wrong-order-option": [
            "search",
            "harry potter",
            "--from",
            "2021-01-01",
            "--order",
            "alphabet",
            "--key",
            API_KEY,
            "--limit",
            "2"
        ],
        "safe-search": [
            "search",
            "harry potter",
            "--from",
            "2021-01-01",
            "--to",
            "2021-01-02",
            "--safe-search",
            "moderate",
            "--key",
            API_KEY,
            "--limit",
            "2"
        ],
        "long-videos": [
            "search",
            "harry potter",
            "--from",
            "2021-01-01",
            "--to",
            "2021-01-02",
            "--safe-search",
            "moderate",
            "--video-duration",
            "long",
            "--key",
            API_KEY,
            "--limit",
            "2"
        ],
        "no-query": ["search", "--from", "2022-08-01", "--key", API_KEY],
    }

    return params


@pytest.mark.parametrize(
    "command", ["wrong-date-format", "wrong-order-option"]
)
def test_cli_search_fail(runner, search_params, command):
    results = runner.invoke(youte, search_params[command])
    assert results.exception


@pytest.mark.parametrize(
    "command",
    ["standard-no-output", "ordered-by-relevance", "safe-search", "long-videos"],
)
def test_cli_search(runner, search_params, command):
    results = runner.invoke(youte, search_params[command])
    assert results.exit_code == 0
    assert "youtube#searchListResponse" in results.output


def test_cli_search_output(runner, tmp_path, search_params):
    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        runner.invoke(youte, search_params["standard-output"])
        assert os.path.exists("output.json")

        with open("output.json", "r") as file:
            row = file.readline()
            r = json.loads(row)

            assert len(r["items"]) > 5


@pytest.fixture()
def related_params() -> dict:
    return {
        "one-id": [
            "get-related",
            "f3m_WqxhL4o",
            "--output",
            "related.json"
        ],
        "multi-ids": [
            "get-related",
            "f3m_WqxhL4o",
            "17yO5AssjAI",
            "--output",
            "related.json"
        ],
        "from-file": [
            "get-related",
            "-f",
            (Path("tests") / "related_ids.csv").resolve(),
            "--output",
            "related.json"
        ]
    }


@pytest.mark.parametrize(
    "command",
    ["one-id", "multi-ids", "from-file"]
)
def test_cli_related_search(runner, tmp_path, related_params, command):
    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        runner.invoke(youte, related_params[command])
        assert os.path.exists("related.json")

        with open("related.json", "r") as file:
            row = file.readline()
            r = json.loads(row)

            assert len(r["items"]) > 10


@pytest.mark.parametrize(
    "country",
    [None, "au", "vn"]
)
def test_cli_most_popular(runner, tmp_path, country):
    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        runner.invoke(youte, ["most-popular", "test.json", "-r", country])
        assert os.path.exists("test.json")

        with open("test.json", "r") as file:
            row = file.readline()
            r = json.loads(row)

            assert len(r["items"]) > 40



import os
import pytest
from click.testing import CliRunner

from youte.cli import youte


@pytest.fixture()
def runner() -> CliRunner:
    runner = CliRunner()
    yield runner
    try:
        os.remove("youte.log")
    except FileNotFoundError:
        pass


@pytest.mark.parametrize(
    "file",
    [
        "tests/search_results.json",
        "tests/videos.json",
        "tests/channels.json",
        "tests/comments.json",
    ],
    ids=["search", "videos", "channels", "comments"],
)
def test_tidy(runner, tmp_path, file):
    results = runner.invoke(youte, ["tidy", file, str(tmp_path / "output.db")])
    assert results.exit_code == 0
    assert (tmp_path / "output.db").exists()



import logging
import configobj

from youte.exceptions import ValueAlreadyExists

logger = logging.getLogger(__name__)


class YouteConfig(configobj.ConfigObj):
    def __init__(self, filename):
        super(YouteConfig, self).__init__(filename)
        self.file_path = filename

    def add_profile(self, name, key):
        existing = []

        try:
            for profile in self:
                existing.append(self[profile]["key"])
        except ValueError:
            pass

        if key in existing:
            raise ValueAlreadyExists("API key already exists.")

        self[name] = {}
        self[name]["key"] = key
        self.write()
        logger.info(f"Add key to config file {self.filename}.")

    def delete_profile(self, name):
        try:
            del self[name]
            self.write()
        except KeyError:
            raise KeyError(f"{name} not found.")

    def set_default(self, name):
        for profile in self:
            if "default" in self[profile]:
                del self[profile]["default"]

        try:
            self[name]["default"] = True
            self.write()
            logger.info("%s is now the default API key." % name)
        except KeyError:
            raise KeyError(f"{name} not found.")



class InvalidFileName(Exception):
    pass
    # def __str__(self):
    #     return "Incorrect file extension or file does not exist."


class InvalidRequestParameter(Exception):
    pass
    # def __str__(self):
    #     return "Paramater fields are invalid for this request."


class StopCollector(Exception):
    pass


class ValueAlreadyExists(Exception):
    pass



import json
import logging
import os
import random
import sqlite3
import sys
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Mapping, Tuple, Union

import click
import requests
from dateutil import tz
from tqdm import tqdm

from youte.exceptions import StopCollector
from youte.utilities import create_utc_datetime_string

logger = logging.getLogger(__name__)


class ProgressSaver:
    def __init__(self, path):
        self.path = path
        self.conn = sqlite3.connect(path, isolation_level=None)

    def load_token(self) -> List:
        self.conn.executescript(
            """
            CREATE TABLE IF NOT EXISTS history (
                next_page_token PRIMARY KEY,
                retrieval_time
                );

            CREATE TABLE IF NOT EXISTS meta (
                id INT CHECK (id=1) UNIQUE,
                params
                );

            INSERT OR IGNORE INTO history(next_page_token)
            VALUES ("");
            """
        )

        tokens = [
            row[0]
            for row in self.conn.execute(
                """
                SELECT next_page_token FROM history
                WHERE retrieval_time IS NULL
                """
            )
        ]

        return tokens

    def update_token(self, token: str) -> None:
        self.conn.execute(
            "REPLACE INTO history VALUES (?,?)", (token, datetime.now(tz=tz.UTC))
        )
        self.conn.commit()

    def add_token(self, token: str) -> None:
        self.conn.execute(
            """
            INSERT OR IGNORE INTO history(next_page_token)
            VALUES (?)
            """,
            (token,),
        )
        self.conn.commit()

    def add_meta(self, params: Mapping) -> None:
        self.conn.execute(
            """
            INSERT OR IGNORE INTO meta
            VALUES (?, ?)
            """,
            (1, json.dumps(params)),
        )

    def get_meta(self) -> Mapping:
        params = [row[0] for row in self.conn.execute("SELECT params FROM meta")]
        if len(params) > 1:
            logger.error("Something is seriously wrong!")
        else:
            return json.loads(params[0])

    def close(self) -> None:
        self.conn.close()


def _load_page_token(filename) -> Tuple[List, ProgressSaver]:
    history = ProgressSaver(path=filename)
    tokens = history.load_token()

    return tokens, history


def _request_with_error_handling(url: str, params: Mapping) -> requests.Response:
    response = requests.get(url, params=params)
    logger.info(f"Getting {response.url}.\nStatus: {response.status_code}.")

    if response.status_code in [403, 400, 404]:
        errors = response.json()["error"]["errors"]

        logger.error(f"Error: {response.url}")
        logger.error(json.dumps(errors))

        for error in errors:
            # ignore if comment is disabled
            if error["reason"] == "commentsDisabled":
                logger.warning(error["reason"])
            elif "quotaExceeded" in error["reason"]:
                # wait until reset time if quota exceeded
                until_reset = _get_reset_remaining(datetime.now(tz=tz.UTC))

                logger.error(error["reason"])
                click.echo(error["reason"])
                logger.warning(f"Sleeping for {until_reset} seconds til reset time")
                click.echo(f"Sleeping for {until_reset} seconds til reset time")
                time.sleep(until_reset)
                _request_with_error_handling(url, **params)
            else:
                logger.error(f"Reason: {error['message']}")
                logger.error(json.dumps(response.json()))
                click.echo(f"{error['message']}")
                sys.exit(1)

    return response


def _write_output_to_jsonl(output_path: str, json_rec: Mapping) -> None:
    logger.info(f"Writing output to {output_path}.")
    with open(output_path, "a") as file:
        json_record = json.dumps(json_rec)
        file.write(json_record + "\n")


def _get_params(**kwargs) -> dict:
    """Populate request parameters"""
    params = {}

    for key, value in kwargs.items():
        # proper datetime string formatting
        if key == "publishedAfter" or key == "publishedBefore":
            value = create_utc_datetime_string(value)
        params[key] = value

    return params


def _prompt_save_progress(filename) -> None:
    if os.path.exists(filename):
        if click.confirm("Do you want to save your current progress?"):
            full_path = Path(filename).resolve()
            click.echo(f"Progress saved at {full_path}")
        else:
            os.remove(filename)
    sys.exit()


def _get_page_token(response: requests.Response, saved_to: ProgressSaver) -> None:
    try:
        next_page_token = response.json()["nextPageToken"]
        saved_to.add_token(next_page_token)
    except KeyError:
        logger.info("No more page...")


def _get_history_path(outfile: Union[str, Path]) -> Path:
    history_dir = Path(".youte.history")

    if not history_dir.exists():
        os.mkdir(history_dir)

    db_file = Path(outfile).with_suffix(".db")

    return history_dir / db_file.name


def _get_reset_remaining(current: datetime) -> int:
    next_reset = datetime.now(tz=tz.gettz("US/Pacific")) + timedelta(days=1)
    next_reset = next_reset.replace(hour=0, minute=0, second=0, microsecond=0)
    reset_remaining = next_reset - current

    return reset_remaining.seconds + 1


class Youte:
    def __init__(self, api_key: str):
        self.api_key = api_key
        self.history_file = None

    def search(
        self, save_progress_to: Union[str, Path] = None, limit: int = None, **kwargs
    ) -> dict:
        page = 0

        if not save_progress_to:
            save_progress_to = f"search_{int(time.time())}.db"

        history_file = _get_history_path(save_progress_to)
        self.history_file = history_file

        try:
            while True:
                tokens, history = _load_page_token(history_file)

                history.add_meta(kwargs)

                if not tokens:
                    break

                for token in tokens:
                    page += 1

                    if limit and page > limit:
                        if history_file.exists():
                            os.remove(history_file)
                        sys.exit(0)

                    url = "https://www.googleapis.com/youtube/v3/search"

                    params = _get_params(key=self.api_key, **kwargs)

                    if token != "":
                        params["pageToken"] = token

                    click.echo(f"Getting page {page}")

                    r = _request_with_error_handling(url=url, params=params)

                    _get_page_token(response=r, saved_to=history)

                    # store recorded and unrecorded tokens in config file
                    history.update_token(token)

                    yield r.json()

            history.close()

            if history_file.exists():
                os.remove(history_file)

        except KeyboardInterrupt:
            raise StopCollector()

        finally:
            if history_file.exists():
                history.close()

    def list_items(
        self,
        item_type,
        ids: List,
        by=None,
        save_progress_to: Union[str, Path] = "history.db",
        **kwargs,
    ):
        valid_type = ["videos", "channels", "comments", "comment_threads"]

        if item_type not in valid_type:
            raise ValueError(f"`item_type` must be one of {valid_type}")

        request_info = _get_endpoint(item_type)
        url = request_info["url"]
        params = request_info["params"]
        params["key"] = self.api_key

        for key, value in kwargs.items():
            params[key] = value

        ids = list(set(ids))

        # SET COLLECTOR TYPE
        # YouTube endpoints fall into 2 categories:
        # 1. A comma-separated list of ids can be passed in the request param
        # 2. Only a single id can be passed in the request param

        valid_by_values = ["parent", "video"]

        if by and by not in valid_by_values:
            raise ValueError("`by` must be `parent` or `video`")

        not_comments = item_type in ["channels", "videos"]
        get_comments_only = (item_type == "comments") and (
            not (by == "parent" or by == "video")
        )
        get_by_parents = (item_type == "comments") and by == "parent"
        get_by_video_channel = (item_type == "comment_threads") and (by == "video")

        can_batch_ids = not_comments or get_comments_only
        cannot_batch_ids = get_by_parents or get_by_video_channel

        if get_comments_only:
            del params["maxResults"]

        if can_batch_ids:
            logger.info("Batching ids")
            click.secho("Getting data")
            while ids:
                if len(ids) <= 50:
                    ids_string = ",".join(ids)
                    ids = None
                else:
                    total_batches = int(len(ids) / 50)
                    click.echo(f"{total_batches} pages remaining")
                    batch = random.sample(ids, k=50)
                    ids_string = ",".join(batch)
                    for elm in batch:
                        ids.remove(elm)

                params["id"] = ids_string
                r = _request_with_error_handling(url=url, params=params)
                yield r.json()

        elif cannot_batch_ids:
            history_file = _get_history_path(save_progress_to)

            for each in tqdm(ids):
                if by == "parent":
                    params["parentId"] = each
                elif by == "video":
                    params["videoId"] = each

                try:
                    while True:
                        if "pageToken" in params:
                            del params["pageToken"]

                        tokens, history = _load_page_token(history_file)

                        if not tokens:
                            logger.info("No more token found.")
                            break

                        for token in tokens:
                            if token != "":
                                logger.info("Adding page token...")
                                params["pageToken"] = token

                            r = _request_with_error_handling(url=url, params=params)

                            _get_page_token(response=r, saved_to=history)

                            # store recorded and unrecorded tokens
                            history.update_token(token)

                            yield r.json()

                except KeyboardInterrupt:
                    sys.exit(1)

                finally:
                    history.close()
                    if history_file.exists():
                        os.remove(history_file)

    def list_most_popular(
        self,
        region_code: str = None,
        video_category: str = None,
    ) -> Mapping:
        request_info = _get_endpoint("videos")
        url = request_info["url"]
        params = request_info["params"]
        params["key"] = self.api_key
        params["chart"] = "mostPopular"

        if region_code:
            params["regionCode"] = region_code
        elif video_category:
            params["videoCategoryId"] = video_category

        history_file = _get_history_path(f"history_{time.time()}.db")

        try:
            while True:
                if "pageToken" in params:
                    del params["pageToken"]

                tokens, history = _load_page_token(history_file)

                if not tokens:
                    logger.info("No more token found.")
                    break

                for token in tokens:
                    if token != "":
                        logger.info("Adding page token...")
                        params["pageToken"] = token

                    r = _request_with_error_handling(url=url, params=params)

                    _get_page_token(response=r, saved_to=history)

                    # store recorded and unrecorded tokens
                    history.update_token(token)

                    yield r.json()

        except KeyboardInterrupt:
            sys.exit(1)

        finally:
            history.close()
            if history_file.exists():
                os.remove(history_file)


def _get_endpoint(endpoint) -> dict:
    """Get all the request details needed to make an API request to
    specified endpoints.

    Position argument: endpoint -- either 'search', 'video', 'channel',
    'commentThread', or 'comment'
    """
    api_endpoints = {
        "search": {
            "url": r"https://www.googleapis.com/youtube/v3/search",
            "api_cost_unit": 100,
            "params": {
                "part": "snippet",
                "maxResults": 50,
                "q": None,
                "type": "video",
                "order": "date",
                "safeSearch": "none",
                "key": None,
            },
        },
        "videos": {
            "url": r"https://www.googleapis.com/youtube/v3/videos",
            "api_cost_unit": 1,
            "params": {
                "part": (
                    "snippet,statistics,topicDetails,status,"
                    "contentDetails,recordingDetails,id"
                ),
                "id": None,
                "maxResults": 50,
                "key": None,
            },
        },
        "channels": {
            "url": r"https://www.googleapis.com/youtube/v3/channels",
            "api_cost_unit": 1,
            "params": {
                "part": (
                    "snippet,statistics,topicDetails,status,"
                    "contentDetails,brandingSettings,contentOwnerDetails"
                ),
                "id": None,
                "maxResults": 50,
                "key": None,
            },
        },
        "comment_threads": {
            "url": r"https://www.googleapis.com/youtube/v3/commentThreads",
            "api_cost_unit": 1,
            "params": {"part": "snippet", "maxResults": 100, "order": "time"},
        },
        "comments": {
            "url": r"https://www.googleapis.com/youtube/v3/comments",
            "api_cost_unit": 1,
            "params": {"part": "snippet", "maxResults": 100},
        },
    }

    return api_endpoints[endpoint]



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

if __name__ == "__main__":
    setup(
        name="youte",
        author="Digital Observatory",
        authhor_email="digitalobservatory@qut.edu.au",
        version="1.3.0",
        description="Command-line tool to collect video metadata and comments from "
                    "Youtube API",
        long_description=long_description,
        long_description_content_type="text/markdown",
        classifiers=[
            'Development Status :: 4 - Beta',
            'Programming Language :: Python :: 3.8'
            ],
        packages=find_packages(where="src"),
        install_requires=[
            'click >= 8.0.3',
            'requests >= 2.27.1',
            'tqdm >= 4.64.0',
            'python-dateutil >= 2.8.2',
            'configobj >= 5.0.6'
            ],
        python_requires='>=3.8',
        entry_points={
            'console_scripts': [
                'youte=youte.cli:youte'
                ]
            }
    )



import nox

nox.options.sessions = ["lint"]


@nox.session
def lint(session):
    session.install("flake8", "black")
    session.run("flake8", "src/youte/")
    session.run("black", "--check", "src/youte/")


@nox.session(python=["3.8", "3.9", "3.10"])
def test(session):
    session.run("pip", "install", "--upgrade", "pip", "setuptools", "wheel")
    session.install(".")
    session.install("pytest")
    session.run("pytest", "tests")



"""
Copyright: Digital Observatory 2022 <digitalobservatory@qut.edu.au>
Author: Boyd Nguyen <thaihoang.nguyen@qut.edu.au>
"""
import json
import logging
import os
import sys
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import List, Sequence, Union, Iterable

import click

from youte import tidier
from youte.collector import Youte, ProgressSaver
from youte.config import YouteConfig
from youte.exceptions import StopCollector, ValueAlreadyExists
from youte.utilities import validate_date_string

# Logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)

file_handler = logging.FileHandler("youte.log")
file_handler.setLevel(logging.INFO)
file_formatter = logging.Formatter(
    "%(asctime)s - %(module)s: %(message)s (%(levelname)s)"
)
file_handler.setFormatter(file_formatter)

logger.addHandler(file_handler)


def _validate_date(ctx, param, value):
    if value:
        if validate_date_string(value):
            return value
        else:
            raise click.BadParameter("Date not in correct format (YYYY-MM-DD)")


def _get_history_path(outfile: Union[str, Path]) -> Path:
    history_dir = Path(".youte.history")

    if not history_dir.exists():
        os.mkdir(history_dir)

    db_file = Path(outfile).with_suffix(".db")

    return history_dir / db_file.name


# CLI argument set up:
@click.group()
@click.version_option()
def youte():
    """
    Utility to collect and tidy YouTube meta-data and comments via YouTube API.
    Run `youte --help` to get started.
    """
    pass


@youte.command()
@click.argument("query", required=False)
@click.option(
    "-o",
    "--output",
    type=click.File(mode="w"),
    help="Name of json file to store results to",
)
@click.option(
    "--from", "from_", help="Start date (YYYY-MM-DD)", callback=_validate_date
)
@click.option("--to", help="End date (YYYY-MM-DD)", callback=_validate_date)
@click.option(
    "--type",
    "type_",
    default="video",
    help="Type of resource to search for",
    show_default=True,
)
@click.option("--name", help="Specify an API key name added to youte config")
@click.option("--key", help="Specify a YouTube API key")
@click.option(
    "--order",
    type=click.Choice(
        ["date", "rating", "relevance", "title", "videoCount", "viewCount"],
        case_sensitive=False,
    ),
    help="Sort results",
    show_default=True,
    default="date",
)
@click.option(
    "--safe-search",
    type=click.Choice(["none", "moderate", "strict"], case_sensitive=False),
    help="Include or exclude restricted content",
    default="none",
    show_default=True,
)
@click.option(
    "--lang",
    help="Return results most relevant to a language (ISO 639-1 two-letter code)",
)
@click.option(
    "--region",
    default="US",
    help="Return videos viewable in the specified country (ISO 3166-1 alpha-2 code)",
    show_default=True,
)
@click.option(
    "--video-duration",
    type=click.Choice(["any", "long", "medium", "short"]),
    help="Include videos of a certain duration",
)
@click.option(
    "--channel-type",
    type=click.Choice(["any", "show"]),
    help="Restrict search to a particular type of channel",
)
@click.option(
    "--video-type",
    type=click.Choice(["any", "episode", "movie"]),
    help="Search a particular type of videos",
)
@click.option(
    "--caption",
    type=click.Choice(["any", "closedCaption", "none"]),
    help="Filter videos based on if they have captions",
)
@click.option(
    "--definition",
    "--video-definition",
    "video_definition",
    type=click.Choice(["any", "high", "standard"]),
    help="Include videos by definition",
)
@click.option(
    "--dimension",
    "--video-dimension",
    "video_dimension",
    type=click.Choice(["any", "2d", "3d"]),
    help="Search 2D or 3D videos",
)
@click.option(
    "--embeddable",
    "--video-embeddable",
    "video_embeddable",
    type=click.Choice(["any", "true"]),
    help="Search only embeddable videos",
)
@click.option(
    "--license",
    "--video-license",
    "video_license",
    type=click.Choice(["any", "creativeCommon", "youtube"]),
    help="Include videos with a certain license",
)
@click.option(
    "--max-results",
    type=click.IntRange(0, 50),
    help="Maximum number of results returned per page",
    default=50,
    show_default=True,
)
@click.option(
    "--limit",
    "-l",
    type=click.IntRange(1, 13),
    help="Maximum number of result pages to retrieve",
)
@click.option("--resume", help="Resume progress from this file")
@click.option("--to-csv", type=click.Path(), help="Tidy data to CSV file")
def search(
    query: str,
    output: str,
    from_: str,
    to: str,
    name: str,
    key: str,
    order: str,
    video_duration: str,
    lang: str,
    region: str,
    type_: str,
    channel_type: str,
    video_type: str,
    caption: str,
    video_definition: str,
    video_dimension: str,
    video_embeddable: str,
    video_license: str,
    safe_search: str,
    resume: str,
    to_csv: str,
    limit: int,
    max_results: int,
) -> None:
    """Do a YouTube search

    QUERY: search query
    """
    api_key = key if key else _get_api_key(name=name)
    search_collector = _set_up_collector(api_key=api_key)

    params = {
        "part": "snippet",
        "q": query,
        "maxResults": max_results,
        "order": order,
        "safeSearch": safe_search,
        "videoDuration": video_duration,
        "type": type_,
        "channelType": channel_type,
        "videoCaption": caption,
        "videoDefinition": video_definition,
        "videoDimension": video_dimension,
        "videoEmbeddable": video_embeddable,
        "videoLicense": video_license,
        "videoType": video_type,
        "relevanceLanguage": lang,
        "regionCode": region,
    }

    if from_:
        params["publishedAfter"] = from_
    if to:
        params["publishedBefore"] = to

    try:
        if resume:
            if not _get_history_path(resume).exists():
                raise click.BadParameter("No such history file found")

            resume_source = ProgressSaver(_get_history_path(resume))
            params = resume_source.get_meta()

        results = search_collector.search(
            save_progress_to=resume, limit=limit, **params
        )

        click.echo(params)

        for result in results:
            click.echo(json.dumps(result), file=output)

            if to_csv:
                items = result["items"]
                tidier.tidy_to_csv(items=items, output=to_csv, resource_kind="search")

        if output:
            click.echo(f"Results are stored in {output.name}")

    except StopCollector:
        _prompt_save_progress(search_collector.history_file)


@youte.command()
@click.argument("items", nargs=-1, required=False)
@click.option("-o", "--output", type=click.File(mode="w"), required=False)
@click.option("-f", "--file-path", help="Get IDs from file", default=None)
@click.option(
    "-t",
    "--by-thread",
    "by_parent",
    help="Get all replies to a parent comment",
    is_flag=True,
)
@click.option("-v", "--by-video", help="Get all comments for a video ID", is_flag=True)
@click.option("--name", help="Specify an API key name added to youte config")
@click.option("--key", help="Specify a YouTube API key")
@click.option("--to-csv", type=click.Path(), help="Tidy data to CSV file")
def get_comments(
    items: Sequence[str],
    output: str,
    name: str,
    key: str,
    file_path: str,
    by_video: bool,
    by_parent: bool,
    to_csv,
) -> None:
    """
    Get YouTube comments by video IDs or thread IDs

    The IDs should all belong to one type, i.e. either video or comment thread.
    You cannot mix both video AND comment thread IDs in one command.

    OUTPUT: name of JSON file to store output

    ITEMS: ID(s) of item as provided by YouTube
    """
    api_key = key if key else _get_api_key(name=name)
    collector = _set_up_collector(api_key=api_key)

    ids = _get_ids(string=items, file=file_path)

    if by_video and by_parent:
        click.secho(
            "Both video and parent flags are on. Only one is allowed.",
            fg="red",
            bold=True,
        )
        sys.exit(1)

    item_type = "comments" if by_parent else "comment_threads"

    if by_video:
        by = "video"
    elif by_parent:
        by = "parent"
    else:
        by = None

    results = collector.list_items(item_type=item_type, ids=ids, by=by)

    for result in results:
        click.echo(json.dumps(result), file=output)

        if to_csv:
            items = result["items"]
            tidier.tidy_to_csv(items=items, output=to_csv, resource_kind="comments")

    if output:
        click.echo(f"Results are stored in {output.name}")


@youte.command()
@click.argument("items", nargs=-1, required=False)
@click.option("-o", "--output", type=click.File(mode="w"), required=False)
@click.option("-f", "--file-path", help="Get IDs from file", default=None)
@click.option(
    "--kind",
    type=click.Choice(["videos", "channels", "comments"], case_sensitive=False),
    help="Sort results",
    show_default=True,
    default="videos",
)
@click.option("--name", help="Specify an API key name added to youte config")
@click.option("--key", help="Specify a YouTube API key")
@click.option("--to-csv", type=click.Path(), help="Tidy data to CSV file")
def hydrate(
    items: Sequence[str],
    output: str,
    kind: str,
    file_path: str,
    name: str,
    key: str,
    to_csv,
) -> Union[str, None]:
    """Hydrate YouTube resource IDs

    Get all metadata for a list of resource IDs.
    By default, the function hydrates video IDs.

    The IDs should all belong to one type, i.e. either video, channel, or comment.
    For example, you cannot mix both video AND channel IDs in one command.

    OUTPUT: name of JSON file to store output

    ITEMS: ID(s) of items as provided by YouTube
    """
    api_key = key if key else _get_api_key(name=name)
    collector = _set_up_collector(api_key=api_key)

    ids = _get_ids(string=items, file=file_path)

    results = collector.list_items(item_type=kind, ids=ids)

    for result in results:
        click.echo(json.dumps(result), file=output)

        if to_csv:
            items = result["items"]
            tidier.tidy_to_csv(items=items, output=to_csv, resource_kind=kind)

    if output:
        click.echo(f"Results are stored in {output.name}")


@youte.command()
@click.argument("items", nargs=-1, required=False)
@click.option(
    "-f", "--file-path", type=click.Path(), help="Get IDs from file", default=None
)
@click.option(
    "-o",
    "--output",
    type=click.File(mode="w"),
    help="Name of JSON file to store output",
)
@click.option(
    "--safe-search",
    type=click.Choice(["none", "moderate", "strict"], case_sensitive=False),
    help="Include or exclude restricted content",
    default="none",
    show_default=True,
)
@click.option("--name", help="Specify an API key name added to youte config")
@click.option("--key", help="Specify a YouTube API key")
@click.option(
    "--max-results",
    type=click.IntRange(0, 50),
    help="Maximum number of results returned per page",
    default=50,
    show_default=True,
)
@click.option("--to-csv", type=click.Path(), help="Tidy data to CSV file")
def get_related(
    items: Sequence[str],
    output: str,
    safe_search: str,
    to_csv: str,
    file_path: str,
    name: str,
    key: str,
    max_results: int,
):
    """Get videos related to a video

    ITEMS: ID(s) of videos as provided by YouTube
    """
    api_key = key if key else _get_api_key(name=name)
    collector = _set_up_collector(api_key=api_key)

    ids = _get_ids(string=items, file=file_path)

    for id_ in ids:
        params = {
            "part": "snippet",
            "maxResults": max_results,
            "safeSearch": safe_search,
            "type": "video",
            "relatedToVideoId": id_,
        }

        results = collector.search(**params)

        click.echo(params)

        for result in results:
            click.echo(json.dumps(result), file=output)

            if to_csv:
                items = result["items"]
                tidier.tidy_to_csv(items=items, output=to_csv, resource_kind="search")

        if output:
            click.echo(f"Results are stored in {output.name}")


@youte.command()
@click.argument("output", type=click.File(mode="w"), required=False)
@click.option(
    "-r", "--region-code", help="ISO 3166-1 alpha-2 country codes to retrieve videos"
)
@click.option("--name", help="Specify an API key name added to youte config")
@click.option("--key", help="Specify a YouTube API key")
@click.option("--to-csv", type=click.Path(), help="Tidy data to CSV file")
def most_popular(region_code, output, name, key, to_csv):
    """Return the most popular videos for a region and video category

    By default, if nothing is else is provided, the command retrieves the most
    popular videos in the US.

    OUTPUT: name of JSON file to store results
    """

    api_key = key if key else _get_api_key(name=name)
    collector = _set_up_collector(api_key=api_key)

    results = collector.list_most_popular(region_code=region_code)

    for result in results:
        click.echo(json.dumps(result), file=output)

        if to_csv:
            items = result["items"]
            tidier.tidy_to_csv(items=items, output=to_csv, resource_kind="videos")

    if output:
        click.echo(f"Results are stored in {output.name}")


@youte.command()
@click.argument("infile", type=click.Path())
@click.option(
    "-o", "--output", type=click.File(mode="w"), help="Output text file to store IDs in"
)
def dehydrate(infile, output: str) -> None:
    """Extract an ID list from a file of YouTube resources

    INFILE: JSON file of YouTube resources
    """
    try:
        items = tidier.get_items(infile)
        ids = tidier.get_id(items)
        for id_ in ids:
            click.echo(id_, file=output)
    except JSONDecodeError:
        raise click.BadParameter("File is not JSON.")


@youte.command()
@click.argument("filepath", nargs=-1, type=click.Path(), required=True)
@click.argument("output", type=click.Path(), required=True)
def tidy(filepath, output):
    """Tidy raw JSON response into relational SQLite databases"""
    for file in filepath:
        tidier.master_tidy(filepath=file, output=output)


@youte.command()
@click.option("-v", "--verbose", is_flag=True, help="Display full search info")
@click.option("-c", "--clear", help="Clear searches by ID")
@click.option("-C", "--clear-all", is_flag=True, help="Clear all search IDs")
def history(verbose: bool = False, clear: str = None, clear_all: bool = False):
    """List resumable search IDs"""
    if not os.path.exists(".youte.history"):
        click.echo("No history file found")

    files = os.listdir(".youte.history")
    if not files:
        click.echo("No history file found")
    else:
        if clear_all:
            for file in files:
                try:
                    os.remove(f".youte.history/{file}")
                except FileNotFoundError:
                    pass
            sys.exit(0)
        if clear:
            try:
                os.remove(f".youte.history/{clear}.db")
                sys.exit(0)
            except FileNotFoundError:
                raise click.ClickException(f"No search found for {clear}")
        if not verbose:
            click.echo("\nsearch ID")
            click.echo("--" * 20)
            for file in files:
                click.echo(file.replace(".db", ""))
        else:
            click.echo_via_pager(_generate_meta_from_search_id(files))


@youte.group()
def config():
    """
    Set up API keys to access data from YouTube API
    """


@config.command()
def add_key():
    """
    Add YouTube API key
    """
    click.echo()
    click.secho("Welcome to youte!", fg="green", bold=True)
    click.echo("Configuring your profile...")
    click.echo()
    click.echo("To get started, an API key is required to get data from Youtube API.")
    click.echo()
    click.echo("To obtain an API key, follow the steps in")
    click.echo("https://developers.google.com/youtube/v3/getting-started")
    click.echo()
    click.echo("Once you have an API key, run `youte configure add-key` to start.")

    click.secho("Setting up you Youtube configuration...", fg="magenta")
    click.echo()
    api_key = click.prompt("Enter your API key")
    username = click.prompt("What would you name this key?")

    config_file_path = Path(_get_config_path())

    if not config_file_path.parent.exists():
        click.echo()
        click.secho(
            "Creating config folder at %s" % config_file_path.parent, fg="magenta"
        )
        config_file_path.parent.mkdir(parents=True)

    conf = _set_up_config()
    click.echo(f"Config file is stored at {config_file_path.resolve()}.")

    try:
        conf.add_profile(name=username, key=api_key)
    except ValueAlreadyExists:
        raise click.BadArgumentUsage("API key already exists in config file.")

    click.echo()
    if click.confirm("Set this API key as default?"):
        conf.set_default(username)
    click.echo()
    click.secho("API key successfully configured!", fg="green", bold=True)
    click.echo()
    click.echo("To add more API keys, rerun `youte config add-key`.")
    click.echo("To set a default key, run `youte config set-default <name-of-key>`")


@config.command()
@click.argument("name")
def set_default(name):
    """Set default API key"""
    click.echo("Setting %s as default key" % name)
    conf = _set_up_config()
    try:
        conf.set_default(name)
        click.echo("%s is now your default API key" % conf[name]["key"])
    except KeyError:
        click.secho(
            """No such name found.
            You might not have added this profile or
            added it under a different name.
            Run:
            - `youte config list-keys` to see the list of registered keys.
            - `youte config add-key to add a new API key.
            """,
            fg="red",
            bold=True,
        )


@config.command()
@click.argument("name")
def remove(name):
    """Delete stored API key"""
    conf = _set_up_config()
    conf.delete_profile(name=name)
    click.echo(f"Profile `{name}` successfully removed!")


@config.command()
def list_keys():
    """Show a list of keys already added"""
    click.echo()
    conf = _set_up_config()

    if not Path(_get_config_path()).exists():
        click.echo("No API key has been added to config file.")
    else:
        for name in conf:
            if "default" in conf[name]:
                click.secho("%s -------- %s (*)" % (name, conf[name]["key"]), fg="blue")
            else:
                click.echo("%s -------- %s" % (name, conf[name]["key"]))
        click.echo()
        click.secho(
            "All API keys are stored in %s" % _get_config_path(), fg="green", bold=True
        )
        click.echo()


def _set_up_collector(api_key: str) -> Youte:
    collector = Youte(api_key=api_key)

    return collector


def _get_config_path(filename="config"):
    config_file_path = Path(click.get_app_dir("youte")).joinpath(filename)
    return str(config_file_path)


def _set_up_config(filename=_get_config_path()) -> YouteConfig:
    config_file = YouteConfig(filename=filename)

    return config_file


def _get_ids(string: Sequence[str] = None, file: str = None) -> List[str]:
    _raise_no_item_error(items=string, file_path=file)

    if string:
        ids = list(string)

    if file:
        with open(file, mode="r") as f:
            ids = [row.rstrip() for row in f.readlines()]

    return ids


def _raise_no_item_error(items: Sequence[str], file_path: str) -> None:
    if not (items or file_path):
        click.secho(
            "No item ID is specified. Pass some item IDs or specify a file.",
            fg="red",
            bold=True,
        )
        sys.exit(1)


def _prompt_save_progress(filename) -> None:
    if click.confirm("Do you want to save your current progress?"):
        full_path = Path(filename).resolve()
        click.echo(f"Progress saved at {full_path}")
        click.echo(f"To resume progress, run `search --resume {full_path.stem}`")
    else:
        try:
            os.remove(filename)
        except FileNotFoundError:
            pass
    sys.exit(0)


def _get_api_key(name=None, filename="config"):
    """Get API key from config file.
    If no name is given, use default API key
    """
    logger.info("Getting API key from config file.")
    config_file_path = Path(click.get_app_dir("youte")).joinpath(filename)
    config = YouteConfig(filename=str(config_file_path))

    if name:
        try:
            api_key = config[name]["key"]
        except KeyError:
            click.secho("ERROR", fg="red", bold=True)
            click.secho(
                "No API key found for %s. Did you use a different name?\n"
                "Try:\n"
                "`youte config list-keys` to get a "
                "full list of registered API keys "
                "or `youte config add-key` to add a new API key" % name,
                fg="red",
                bold=True,
            )
            sys.exit(1)
    else:
        default_check = []
        for name in config:
            if "default" in config[name]:
                default_check.append(name)
        if not default_check:
            click.secho(
                "No API key name was specified, and you haven't got a default API key.",
                fg="red",
                bold=True,
            )
            sys.exit(1)
        else:
            api_key = config[default_check[0]]["key"]

    return api_key


def _generate_meta_from_search_id(files: Iterable[str]) -> Sequence:
    for file in files:
        meta = ProgressSaver(_get_history_path(file)).get_meta()
        yield f"Search ID: {file.replace('.db', '')}\n{json.dumps(meta)}\n--------------------------------------------------------\n"
